#ifndef INTERFACE_H
#define INTERFACE_H

#include <stdarg.h>
#include "Map.h"

//extern int ColorMap[22];
void initGlyph (shGlyph *g, char sym, shColor color);

extern int GodMode;


struct shMenu;
struct shTextViewer;
struct shNameToKey
{
    const char *name;
    int key;
};
struct shInterface
{
    enum Command {
        kNoCommand,
        kEscape,
        kSpace,
        kEnter,

        kMoveN,
        kMoveNE,
        kMoveE,
        kMoveSE,
        kMoveS,
        kMoveSW,
        kMoveW,
        kMoveNW,
        kMoveDown,
        kMoveUp,

        kGlideN,
        kGlideNE,
        kGlideE,
        kGlideSE,
        kGlideS,
        kGlideSW,
        kGlideW,
        kGlideNW,
        kGlideDown,
        kGlideUp,

        kFireN,
        kFireNE,
        kFireE,
        kFireSE,
        kFireS,
        kFireSW,
        kFireW,
        kFireNW,
        kFireDown,
        kFireUp,

        kAdjust,
        kClose,
        kDiscoveries,
        kDrop,
        kDropMany,
        kEditSkills,
        kExamine,
        kExecute,
        kFireWeapon,
        kGlide,
        kHelp,
        kHistory,
        kKick,
        kListInventory,
        kLook,
        kMutantPower,
        kName,
        kOpen,
        kPay,
        kPickup,
        kPickUpAll,
        kQuaff,
        kRest,
        kTakeOff,
        kToggleAutopickup,
        kThrow,
        kUse,
        kWear,
        kWield,
        kZapRayGun,

/* debug commands: */

        kGodMode,

        kMaxCommand
    };

    static shDirection
    moveToDirection (Command cmd)
    {
        if (cmd < kMoveN || cmd > kMoveUp) {
            return kNoDirection;
        } else {
            return (shDirection) (cmd - kMoveN);
        }
    }

    //virtual shInterface ();
    virtual ~shInterface () = 0;

    virtual int getChar () = 0;
    virtual int getStr (char *buf, int len, const char *prompt,
                const char *dflt = NULL) = 0;
    int getSquare (const char *prompt, int *x, int *y, int maxradius,
                   int instant = 0);
    Command getCommand ();
    virtual const char *getKeyForCommand (Command cmd) = 0;
    shDirection getDirection (int *x = NULL, int *y = NULL,
                              int *z = NULL, int silent = 0);

    void cursorOnHero ();
    void pauseXY (int x, int y, int ms = 0, int draw = 1);
    virtual void drawSideWin () = 0;
    virtual void drawScreen () = 0;
    virtual void drawLog () = 0;
    virtual void refreshScreen () = 0;

    /* Draws everything specified unquestionably. */
    virtual void draw (int x, int y, shSpecialEffect e, shCreature *c,
               shObjectIlk *o, shFeature *f, shSquare *s) = 0;
    /* Draws everything specified. What is not given is pulled from
       hero's memory of square contents. */
    virtual void drawMem (int x, int y, shSpecialEffect e, shCreature *c,
                  shObjectIlk *o, shFeature *f, shSquare *s) = 0;
    void drawMem (int x, int y) {
        drawMem (x, y, kNone, NULL, NULL, NULL, NULL);
    };


    int yn (const char *prompt, ...); /* Yes or no prompt. */
    int p (const char *format, ...);
    int vp (const char *format, va_list ap);
    /* Diagnostic message output to log and diag window if present. */
    virtual int diag (const char *format, ...) = 0;
    /* Lines that can be printed until "more" appears. */
    int freeLogLines ();
    void nonlOnce ();
    void nevermind ();
    void pause ();
    void smallPause ();
    virtual void pageLog () = 0;
    virtual void showHistory () = 0;
    virtual void showKills () = 0;
    virtual void showHelp () = 0;
    virtual void showVersion () = 0;
    virtual void editOptions () = 0;
    void loadOptions ();
    void saveOptions ();
    virtual void keymapChoice () = 0;
    void doMorePrompt ();
    virtual void doScreenShot (int fileHandle) = 0;
    virtual void showHighScores () = 0;
    void readKeybindings (const char *fname, int n, const shNameToKey table[]);

    virtual void setColor (shColor c) = 0;// { mColor = c; };

    virtual shMenu *newMenu (const char *prompt, int flags) = 0;
    virtual shTextViewer *newViewer (const char *fname) = 0;

 protected:
    /* Functions that isolate whatever screen and keyboard handling
       each inheriting class employ. */

    virtual void cursorOnXY (int x, int y) = 0;
    virtual void outChar (char ch) = 0;
    virtual void outStr (const char *str) = 0;
    virtual void moreBegin () = 0;
    virtual void moreEnd () = 0;
    virtual void logPrint (const char *str) = 0;
    virtual Command keyToCommand (int key) = 0;
    virtual void readKeybindings (const char *fname) = 0;
    virtual void assign (int key, Command cmd) = 0;
    virtual void resetKeybindings () = 0;

    /* Implemented internal helper functions. */
    void delay (int ms);

    /* Variables need for stuff generic interface implements. */
#define HISTORY_ROWS 20
    char mLogHistory[HISTORY_ROWS * 80];
    int mHistoryIdx;
    int mHistoryWrapped;
    int mNoNewline;
    int mPause;

    int mLogRow;      //first empty row in LogWin
    int mLogCol;      // --  || --  col in LogWin
    int mLogSize;     //num rows in LogWin
    int mLogSCount;   //num rows output since last keyboard input
};
inline shInterface::~shInterface (void) {}


/* Value used for determining player's choice to show full inventory
   instead of a subset.  Examples: wielding items, throwing items. */
#define DELETE_FILTER_SIGNAL -1
struct shMenu
{
 public:
    enum MenuFlags {
        kNoPick = 0x1,      /* display only */
        kMultiPick = 0x2,
        kCountAllowed = 0x4,
        kAnythingAllowed = 0x8,
        kNothingAllowed = 0x10, /* for wielding */
        kCategorizeObjects = 0x20, /* hack */
        kNoHelp = 0x40,
        kShowCount = 0x80, /* for skills */
        kSelectIsPlusOne = 0x100 /* for skills */
    };

    virtual ~shMenu () = 0;

    virtual void addPtrItem (char letter, const char *text, const void *value,
                     int count = 1, int selected = 0) = 0;
    void addText (const char *text) { addPtrItem (' ', text, NULL); }
    virtual void addIntItem (char letter, const char *text, const int value,
                     int count = 1, int selected = 0) = 0;
    virtual void addHeader (const char *text) = 0;
    virtual void addPageBreak () = 0;
    virtual int getPtrResult (const void **value, int *count = NULL) = 0;
    virtual int getIntResult (int *value, int *count = NULL) = 0;
    virtual void dropResults () = 0;
    virtual void attachHelp (const char *fname) = 0;
    virtual void finish () = 0;
};
inline shMenu::~shMenu (void) {}



struct shTextViewer
{
    virtual ~shTextViewer (void) = 0;
    virtual void show (void) = 0;
};
inline shTextViewer::~shTextViewer (void) {}

extern shInterface *I;
#endif
